/*
 * Program:	VMS mail routines
 *
 * Author:	Yehavi Bourvine, The Hebrew University of Jerusalem.
 *		Internet: Yehavi@VMS.huji.ac.il
 *
 * Modified:	Andy Harper, Kings College London
 *
 * Date:	2 August 1994
 * Last Edited:	18 November 1996
 *
 * Copyright 1993 by the University of Washington
 *
 *  Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appears in all copies and that both the
 * above copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the University of Washington not be
 * used in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  This software is made
 * available "as is", and
 * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
 * NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
 * (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <types.h>
#include <errno.h>
/* extern int errno;		/* just in case */
#include <signal.h>
#include "mail.h"
#include "osdep.h"
#include <sys/file.h>
#include <sys/stat.h>
/* #include <sys/time.h>		/* must be before osdep.h */
#include "vms_mail.h"
#include "rfc822.h"
#include "misc.h"
#include "dummy.h"
#include <ctype.h>

/*
   Define VMS specific sizes
*/
#define FOLDERSIZE    64		/* Length of a folder name         */
#define MAX_FOLDERS   256		/* max # of folders in a mail file */
#define MAILFILESIZE  256		/* length of a mail file name      */
#define MAXRECORDSIZE 256		/* Max length of a message record  */



/* VMS/mail routines */

/*
   Context Values for use with callable mail

   we also keep the number of messages when the folder was opened. This is
   needed since PINE checks this number once in a while; if new message arrive
   it will fool our cache mechanism.
*/

static int  MailContext = 0;			/* Currently opened VMS/MAIL file */
static int  MessageContext = 0;			/* Currently opened folder */
static int  TotalMessagesCount = -1;		/* How many messages in this folder? */
static char GlobalMailFile[MAILFILESIZE];	/* See description in __vms_mail_open() */


						/* As above, but saved copies of previous folder */
#define MAXSAVEPREV 5
static int   PrevSP = 0;			/* Stack pointer for saves */

static int   PrevMailContext[MAXSAVEPREV];
static int   PrevMessageContext[MAXSAVEPREV];
static int   PrevTotalMessagesCount[MAXSAVEPREV];
static struct message_text_cache *PrevMessageCache[MAXSAVEPREV];
static char  PrevGlobalMailFile[MAXSAVEPREV][MAILFILESIZE];


/* List of folders so we won't open a folder that does not exist */
char	FoldersNames[MAX_FOLDERS][FOLDERSIZE];
int	FoldersNamesCount;

/* Driver dispatch used by MAIL */

DRIVER vmsmaildriver = {
  "vms_mail",			/* driver name */
  (DRIVER *) NIL,		/* next driver */
  vms_mail_valid,		/* mailbox is valid for us */
  vms_mail_parameters,		/* manipulate parameters */
  vms_mail_find,		/* find mailboxes */
  vms_mail_find_bboards,	/* find bboards */
  vms_mail_find_all,		/* find all mailboxes */
  vms_mail_find_all_bboards,	/* find all bboards */
  vms_mail_subscribe,		/* subscribe to mailbox */
  vms_mail_unsubscribe,		/* unsubscribe from mailbox */
  vms_mail_subscribe_bboard,	/* subscribe to bboard */
  vms_mail_unsubscribe_bboard,	/* unsubscribe from bboard */
  vms_mail_create,		/* create mailbox */
  vms_mail_delete,		/* delete mailbox */
  vms_mail_rename,		/* rename mailbox */
  vms_mail_open,		/* open mailbox */
  vms_mail_close,		/* close mailbox */
  vms_mail_fetchfast,		/* fetch message "fast" attributes */
  vms_mail_fetchflags,		/* fetch message flags */
  vms_mail_fetchstructure,	/* fetch message envelopes */
  vms_mail_fetchheader,		/* fetch message header only */
  vms_mail_fetchtext,		/* fetch message body only */
  vms_mail_fetchbody,		/* fetch message body section */
  vms_mail_setflag,		/* set message flag */
  vms_mail_clearflag,		/* clear message flag */
  vms_mail_search,		/* search for message based on criteria */
  vms_mail_ping,		/* ping mailbox to see if still alive */
  vms_mail_check,		/* check for new messages */
  vms_mail_expunge,		/* expunge deleted messages */
  vms_mail_copy,		/* copy messages to another mailbox */
  vms_mail_move,		/* move messages to another mailbox */
  vms_mail_append,		/* append string message to mailbox */
  vms_mail_gc			/* garbage collect stream */
};

				/* prototype stream */
MAILSTREAM vmsmailproto = {&vmsmaildriver};
                                /* standard driver's prototype */
MAILSTREAM *mailstd_proto = &vmsmailproto;

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static struct ITEM_LIST NullItemList[] = {{0, 0, NULL, NULL}};
static struct ITEM_LIST NullItem[]     =  {0, 0, NULL, NULL} ;

int mail$user_begin();
int mail$user_get_info();
int mail$user_set_info();
int mail$user_end();

int mail$message_begin();
int mail$message_select();
int mail$message_get();
int mail$message_delete();
int mail$message_copy();
int mail$message_end();

int mail$mailfile_begin();
int mail$mailfile_open();
int mail$mailfile_info_file();
int mail$mailfile_close();
int mail$mailfile_end();

int mail$send_begin();
int mail$send_add_attribute();
int mail$send_add_address();
int mail$send_add_bodypart();
int mail$send_message();
int mail$send_end();


/*
 ROUTINE CALL/NESTING				PARAMS			CONTEXT
 ====================				======			=======

 mail$mailfile_begin							mail
 :
 :   mail$%mailfile_open			(mailfile)		mail
 :   :
 :   :
 :   :   mail$message_begin						folder
 :   :   :
 :   :   :   mail$message_select		(folder,msgcnt)		folder
 :   :   :
 :   :   :   mail$message_get			(msgno,size)		folder
 :   :   :
 :   :   : ->mail$message_get			(next record)		folder
 :   :   : :     :
 :   :   : ^     V
 :   :   : :     :
 :   :   : +-----+
 :   :   :
 :   :   :   mail$message_copy			(msgno,folder,file)	folder
 :   :   :
 :   :   :   mail$message_delete		(msgno)			folder
 :   :   :
 :   :   mail$message_end						folder
 :   :
 :   :
 :   :   mail$mailfile_info_file		(folder action)		mail
 :   :
 :   :
 :   mail$mailfile_close						mail
 :
 mail$mailfile_end							mail

*/

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purpose:
	Simplify system status checking. Check the status message and print a
	message if it indicates an error. Exit with this status if the abort
        flag is set, otherwise return with the status.

    Author:
 	Andy Harper, Kings College London, OCT 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int chkstat(char *message, int abort, int status)
  {
     if ((status & 0x1) == 0)
       {
          fprintf(stderr, "%s: failed with status = 0x%08x\n", message, status);
          if (abort) exit(status);
       }
     return status;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purpose:
	Get the default MAIL directory for the current user

    Author:
	Andy Harper, Kings College London, OCT 1996

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

char *vms_get_mail_directory(char **Mail)
{
  static char MailDirectory[MAILFILESIZE];	/* Must be static as we return it */
  char username[80],host[256], password[80];
  int length, UserContext, status;

  struct ITEM_LIST InItemList[] = {
	{ 0, MAIL$_USER_USERNAME, username, NULL},
	{ 0, 0, NULL, NULL }};
  struct ITEM_LIST OutItemList[] = {
	{ sizeof(MailDirectory) - 1, MAIL$_USER_FULL_DIRECTORY, MailDirectory, &length},
	{ 0, 0, NULL, NULL }};

  mm_login(host, username, password, -1L);
  InItemList[0].length = strlen(username);

  UserContext = 0;
  status = mail$user_begin(&UserContext, NullItemList, NullItemList);
  if ((chkstat("mail$user_begin", 0, status) & 0x1) == 0) return NULL;
  status = mail$user_get_info(&UserContext, InItemList, OutItemList);
  if ((chkstat("mail$user_get_info", 0, status) & 0x1) == 0) return NULL;
  status = mail$user_end(&UserContext, NullItemList, NullItemList);
  if ((chkstat("mail$user_end", 0, status) & 0x1) == 0) return NULL;

  MailDirectory[length & 0xffff] = '\0';
  if (Mail) *Mail = MailDirectory;
  return MailDirectory;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   Purpose:
	Get the login directory (SYS$LOGIN) for the current user

   Author:
	Andy Harper, Kings College London, 4-OCT-1996

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static char *vms_get_home_directory(char **hp)
  {
     static char home[100];
     static device[32], directory[64];
     int status, sys$getuai();
     char host[256], username[256], password[256];

     struct DESC      usernameD;
     struct ITEM_LIST UserInfo[] = {
	{sizeof(device),       UAI$_DEFDEV,  (char *)device,     0},
	{sizeof(directory),    UAI$_DEFDIR,  (char *)directory,  0},
        {0,                    0,             0,                 0}};

     mm_login(host, username, password, -1L);
     usernameD.address = ucase( username);
     usernameD.length  = strlen(username);
     usernameD.type    = 0;
     status = sys$getuai(0,0,&usernameD, UserInfo,0,0,0);
     if ((chkstat("sys$getuai", 0, status) & 0x1) == 0) return NULL;

     strncpy(home, (char *)device+1, (int) device[0]);
     strncpy(home + (int) device[0], (char *)directory+1, (int) directory[0]);
     home [ ((int) device[0]) + ((int) directory[0])] = '\0';
     *hp = home;
     return home;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   Purpose:
	To parse the supplied mailbox specification, supply defaults where
	necessary (mail directory etc.), and split into a mail file and
	folder specification.

        Mailbox is first uppercased (except where quoted).

   Input Formats:
        INBOX			--> [default mail file], NEWMAIL
	folder			--> [default mail file], folder

      The following for unix: compatibility (only if ALLOWFULLMBXSPEC defined):

	~/folder		--> [relative mail file], folder
        ~/subdir/folder		--> [relative mail file], folder

        The filename part is always "MAIL.MAI" and this should not be
	specified in the pathname.

        A NULL folder name is an error, any other format is invalid

   Output:
       full mail dir spec  (defaulted to current' users default mailfile)
       foldername

   Returns:
  	NIL mailbox parse error
	T   all ok

   Author:
	Andy Harper, Kings College London, 30-SEP-1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static char InboxName[] = "NEWMAIL";

static int vms_parse_mailbox (char *name, char *directory, char *folder)
  {
#ifdef ALLOWFULLMBXSPEC
    static char *invalchrs = "{}()[]#^&'";
#else
    static char *invalchrs = "{}()[]#^&'~";
#endif
    char mailbox[256];
    char c, *i, *p = name, *mbox = mailbox;

    if ( name == NIL )
      return NIL;

						/* Check invalid characters */
    for (i=invalchrs; *i; i++) if (strchr(name,*i)) return NIL;

						/* Uppercase, except where quoted */
    while ( (c = *p++) )
      if (c == '"')
        {
          while ( (c = *p++) && (c != '"') )  *mbox++ = c;
          if ( c == '\0') break;
        }
      else
        *mbox++ =  islower(c) ? toupper(c) : c;

     *mbox = '\0';


						/* Special INBOX handling */
    mbox = mailbox;
    if ( strcmp(mbox, "INBOX") == 0)
      {
         strcpy(directory, vms_get_mail_directory(&p) );
         mbox = InboxName;
      }

#ifdef ALLOWFULLMBXSPEC
						/* handle home-relative spec */
    else if ( *mbox == '~' )
      {
         if (mbox[1] != '/') return NIL;	/* ... must start '~/'  ... */
         mbox = mbox+2;

         strcpy(directory,vms_get_home_directory(&p));
         directory[strlen(directory)-1] = '\0';	/* Strip ']' from end */

						/* strip paths, convert to VMS subdirs */
         while ( (p=strchr(mbox, '/') ) != NULL)
           {
              *p = '\0';			/* Mask out '/'  */
              strcat(directory,".");		/* Add subdir separator */
              strcat(directory,mbox);		/* Add subdir */
              mbox = p+1;			/* start of next part */
           }

         strcat(directory,"]");			/* Terminate directory part ..*/
      }
#endif


						/* Simple folder name */
    else
     {
        if (strchr(mbox,'/') != NULL) return NIL;	/* ... no paths allowed */
        strcpy(directory,vms_get_mail_directory(&p));
     }

    if (*mbox == '\0') return NIL;

						/* Append filename and folder */
    strcat(directory,"MAIL.MAI");
    strcpy(folder,mbox);
    return T;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  Purpose:
    Help Map DECnet style addresses (namespace:.node::user, node::user and user)
    into RFC822 style, using logical names:

      IMAPD_NODE_namespace_node	If both namespace and node   [ namespace:.node::user ]
      IMAPD_NODE_node		If nodename only             [ node::user ]
      IMAPD_NODE		If no namespace or nodename  [ username]

    What we return is either the translation of the appropriate logical name
    the original nodename, or (as a last resort) the current system name, for
    use in constructing the user@node RFC822 format

  Author:
    Andy Harper, Kings College London, October 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static char *vms_map_nodename(char *namespace, char *nodename)
  {
     static char hostname[256];
     char *p, logname[256] = "IMAPD_NODE";
     int c, s = strlen(logname);


					/* Add namespace part to logical name */
     if (namespace && *namespace)
       {
         logname[s++] = '_';
         while (c = *namespace++)
            logname[s++] = islower(c) ? toupper(c) : c;
       }

					/* Add nodename part to logical name */
     if (nodename && *nodename)
       {
         logname[s++] = '_';
         while (c = *nodename++)
            logname[s++] = islower(c) ? toupper(c) : c;
       }
					/* Finish off the string */
     logname[s]   = '\0';

					/* Return any translation */
     if ((p=getenv(logname)) != NULL)
        return p;

					/* .. or the original nodename */
     if (nodename != NULL)
        return nodename;

					/* .. or the local hostname */
     get_local_host_name(hostname,sizeof(hostname));
     return hostname;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Given an address string that resembles one of the standard VMS MAIL
	From: headers, return an RFC822 equivalent.

   Conversions:
	xyz%"..."		-->	...
	abc@def			-->	abd@def
	user			-->	username@node
	node::user		-->	user@node
	namespace:.node::user	-->	user@node

	In the last three cases, the nodename/namespace is optionally mapped
	by logicals into an alternative nodename string

   Author:
	Andy Harper, Kings College London, October 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static char *vms_to_rfc822(char *decnetaddress)
  {
    static char rfc822address[256];
    char *c, *namespace = NULL, *nodename = NULL;


					/* Look for  xyz%"...."     */
    if ((c=strchr(decnetaddress,'%')) != NULL)
      {
        if ( *++c == '"')
           {
             c++;
             if (strrchr(c,'"') != NULL) *strrchr(c,'"') = '\0';
             
           }
        strcpy(rfc822address,c);
        return rfc822address;
      }


					/* Look for internet style addresses */
    if (strchr(decnetaddress,'@') != NULL)
      {
        strcpy(rfc822address,decnetaddress);
        return rfc822address;
      }

					/* Look for DECnet Phase V namespace */
    if ((c=strstr(decnetaddress,":.")) != NULL)
      {
        *c = '\0';
        namespace = decnetaddress;
        decnetaddress = c + 2;
      }

					/* Look for a DECnet nodename */
    if ((c=strstr(decnetaddress,"::")) != NULL)
      {
         *c = '\0';
         nodename = decnetaddress;
         decnetaddress = c + 2;
      }

					/* Convert to RFC822 */
    strcpy(rfc822address,decnetaddress);
    strcat(rfc822address,"@");
    strcat(rfc822address, vms_map_nodename(namespace, nodename));

    return rfc822address;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	VMS's DECnet style addressing will confuse things, so we'll reformat
	them into a more RFC822-like syntax here.

   Input formats:
        user				(no DECnet)
        node::user			(DECnet IV)
        namespace:.node::user		(DECnet V)

        plus, comma separated list of the above

   Output formats:
	user@nodename

   Note:
	If any translated address will not fit into the output record, it
	is silently dropped. This prevents writing into random memory!

   Author:
	1.0 Andy Harper 15-AUG-1996	First version
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static void vms_header_rewrite(char *record, int maxlen)
 {
   char header[256],temp[256]="";
   char *start,*p,*a;
   int size;

   sscanf(record,"%s %s",header,temp);
   size = strlen(header);

					/* If SUBJ:, replace with SUBJECT: */
   if (search(header, size, "subj:", 5) != NIL)
     {
        strcpy(temp,record);
        strcpy(record,"Subject:");
        strcat(record,temp+5);
        return;
     }

					/* Reformat only address headers */
   if (
       search(header, size, "cc:",   3) == NIL &&
       search(header, size, "to:",   3) == NIL &&
       search(header, size, "from:", 5) == NIL
						)      return;

					/* Rebuild header line */
   strcpy(record,header);
   strcat(record," ");

					/* Scan list of addresses */
   start = temp;
   while (*start)
     {
					/* Last username in list ??? */
       if ((p=strstr(start,",")) == NULL)
        {
           a = vms_to_rfc822(start);
           if (strlen(record) + strlen(a) < maxlen)  strcat(record, a);
           break;
        }

					/* Not last !! */
       *p = '\0';
       a = vms_to_rfc822(start);
       if (strlen(record) + strlen(a) +1 < maxlen)
         {strcat(record, a);
          strcat(record,",");
         }
       start = p + 1;			/* Go to next username listed */
     }
 }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purpose:
	This routine examines the header record and returns T if it looks like
	an RFC822 header and NIL otherwise. Address fields must contain an
	'@' symbol to reduce the risk of lines starting with a valid header
        being mistakenly interpreted.

    Author:
	Andy Harper, Kings College London, Sep 30 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static vms_record_is_rfc822(char *record)
  {
    int   size = strlen(record);
    char *ats  = strchr(record,'@');

    if (  search(record, size, "to:",           3 ) == T  && ats) return T;
    if (  search(record, size, "via:",          4 ) == T        ) return T;
    if (  search(record, size, "date:",         5 ) == T        ) return T;
    if (  search(record, size, "from:",         5 ) == T  && ats) return T;
    if (  search(record, size, "comment:",      8 ) == T        ) return T;
    if (  search(record, size, "received:",     9 ) == T        ) return T;
    if (  search(record, size, "message-id:",   11) == T        ) return T;
    if (  search(record, size, "return-path:",  12) == T  && ats) return T;
    if (  search(record, size, "mime-version:", 13) == T        ) return T;
    if (  search(record, size, "content-type:", 13) == T        ) return T;

    return NIL;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 * Get the next record from the message which should have been selected
 * already. Return the record's size read.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

static int get_vms_message_record(char *record, int size)
  {
    static int RecordLength;	/* Returned by mail reading proc */
    int i, status;
    struct ITEM_LIST GetMessageTextItem[] = {
        { sizeof(int), MAIL$_MESSAGE_CONTINUE, NULL, NULL },
        { 0, 0, NULL, NULL }};
    struct ITEM_LIST RecordItem[] = {
        { 0, MAIL$_MESSAGE_RECORD, 0, &RecordLength },
        { 0, 0, NULL, NULL }};


/*  Get the next line of data from the message */
    RecordItem[0].length = size;
    RecordItem[0].buffer = record;
    status = mail$message_get(&MessageContext, GetMessageTextItem, RecordItem);
    if((status & 0x1) == 0)		/* Probably end of message */
      {
        record[0] = '\0';		/* return a valid null record */
        return -1;
      }
    record[RecordLength] = '\0';


/* Look for nulls inside the record (can't be inside C's strings...). Replace
   them with spaces. */
    for(i = 0; i < RecordLength; i++)
      if(record[i] == '\0') record[i] = ' ';

    return RecordLength;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Purpose:
	Get the delivery date of a message

 Author:
	Andy Harper, Kings College London, November 1996

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static char *vms_mail_datefield (long m)
{
  static int  datelen;
  static char dateinfo[256];
  int	status;
  struct ITEM_LIST GetMessageInfoItem[] = {
	{ sizeof(m), MAIL$_MESSAGE_ID, (char *)&m, NULL },
	{ 0, 0, NULL, NULL }};
  struct ITEM_LIST AnswerMessageInfoItem[] = {
	{ sizeof(dateinfo)-1,  MAIL$_MESSAGE_DATE, dateinfo, &datelen },
	{ 0, 0, NULL, NULL }};

  status = mail$message_get(&MessageContext, GetMessageInfoItem, AnswerMessageInfoItem);
  chkstat("mail$message_get", 1, status);

  dateinfo[datelen & 0xffff] = '\0';
  return dateinfo;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  Purpose:
    The message is read, line by line, from the mail file and, under VMS, has
    this format:
         VMS MAIL headers (From, To, CC, Subj)
         Blank Line
         [Optional RFC822 headers, if delivered by a network transport,
          Blank line]
         Body of text

    This routine reads the message and separates it into headers
    and text parts. Where RFC822 headers are (apparently) present, they
    overwrite the VMS MAIL headers, otherwise the VMS MAIL headers are
    reformatted into an rfc822-like syntax where possible.

    Andy Harper, Kings College London, Sep 30 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static void vms_split_message (MAILSTREAM *stream, long msgno, unsigned long int size,
                       char **hdr,  unsigned long int *hdrsize,
                       char **text, unsigned long int *textsize)
  {
     char record[MAXRECORDSIZE], temp[MAXRECORDSIZE], *mbuf;
     int CurrentSize, len, record_s = sizeof(record);
     char *deliverydate = vms_mail_datefield(msgno);

     
     mbuf  = (char *) fs_get(size);
     *mbuf = *temp = '\0';

  /* Read the initial VMS headers into the buffer */
     while ( get_vms_message_record(record, record_s) > 0 )
       {
          vms_header_rewrite(record,sizeof(record));	/* Reformat addresses */
          strcat(mbuf,record);
          strcat(mbuf,"\n");
       }

 /* Terminate VMS mail headers with an extra delivery date field */
     strcat(mbuf,"Date: ");  strcat(mbuf, deliverydate); strcat(mbuf,"\n");
     mail_parse_date( mail_elt(stream,msgno), deliverydate);


  /* If the next line LOOKS LIKE an RFC header, then use them instead */
     if ( get_vms_message_record(record, record_s) > 0 )
       {
         if ( vms_record_is_rfc822(record) )
           {
             *mbuf = '\0';
             do {
               strcat(mbuf,record); strcat(mbuf,"\n");
             } while ( get_vms_message_record(record, record_s) > 0 );
           }
         else	/* not RFC  header, must be start of text body, save it .. */
           {
             strcpy(temp,record); strcat(temp,"\n");
           }
       }
     else	/* Blank line is 1st line of text body !! */
       {
         strcpy(temp,"\n");
       }


  /* Save the extracted headers into a new buffer if they are required */
     if (hdr != NULL)
       {
         strcat(mbuf,"\n");			/* terminating blank line */
         *hdrsize = strlen(mbuf);
         *hdr     = (char *) fs_get(*hdrsize + 1);
         strcpy(*hdr, mbuf);
       }


  /* Get the text body if it has been requested */
     if (text != NULL)
       {
         strcpy(mbuf,temp);
         CurrentSize = strlen(mbuf);
         while ( (len = get_vms_message_record(record, record_s)) >= 0 )
           {
             strcpy(&mbuf[CurrentSize], record);
             mbuf[CurrentSize += len] = '\n';
             mbuf[++CurrentSize] = '\0';
           }

         *textsize = CurrentSize;
         *text     = mbuf;
       }
     else
       {
         fs_give((void **) &mbuf);
       }
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purpose:
	If an inbox folder, flag message as seen

    Author:
	Andy Harper, Kings College London, November 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static void vms_flag_seen_if_inbox(MAILSTREAM *stream,int msgno)
  {
    char  mailfile[MAILFILESIZE], folder[FOLDERSIZE], sequence[256];
    vms_parse_mailbox(stream->mailbox, mailfile, folder);
    if(strcmp(folder, InboxName) == 0)
      {
	sprintf(sequence, "%d", msgno);	/* Write message number as string */
	vms_mail_setflag (stream,sequence,"(\\SEEN)");
      }
  }

/* VMS/MAIL mail lock file for parse/append permission
 * Accepts: file descriptor
 *	    lock file name buffer
 *	    type of locking operation (LOCK_SH or LOCK_EX)
 * Returns: file descriptor of lock or -1 if failure
 */

int vms_mail_lock (FILE *fd, char *lock, int op)
{
 return 0;		/* no locking needed */
}


/* VMS/MAIL mail unlock file for parse/append permission
 * Accepts: file descriptor
 *	    lock file name from vms_mail_lock()
 */

void vms_mail_unlock (FILE *fd, char *lock)
{
  return;
}

/*
 | Init the text cache.
 */
static void init_vms_mail_cache(int nmsgs)
  {
    int i;

					/* Allocate the message cache */
    MessageTextCache = fs_get(nmsgs * sizeof(struct message_text_cache));

					/* Clear all entries to null pointers */
    for(i = 0; i < nmsgs; i++) {
	MessageTextCache[i].text = NIL;
	MessageTextCache[i].flags = 0;
    }
  }


/*
 | Clear the text cache.
 */
static void clear_vms_mail_cache(int nmsgs)
  {
    int i;

    if(nmsgs == 0) return;	/* Happens when new mail arrives in the middle of delete */

				/* Free all occupied entries */
    for(i = 0; i < nmsgs; i++)
	if(MessageTextCache[i].text != NIL)
		free(MessageTextCache[i].text);

				/* Free the cache table itself */
    fs_give((void **) &MessageTextCache);
  }

static char *pattern = NULL;

/* this routine is called for each folder by the INFO routine. It reports
   the name of the folder back to PINE.
*/
static int list_folders(int *UserData, struct DESC *FolderName)
  {
    char folder[FOLDERSIZE];

    if(FolderName == NULL) return;	/* Shouldn't happen */
    if(FolderName->length == 0) return;	/* End of list */

    strncpy(folder, FolderName->address, FolderName->length);
    folder[FolderName->length & 0xffff] = '\0';
    mm_mailbox(folder);		/* Register it */
    return 1;
}



/*
 | This routine does much the same as the previous one, but this time
 | it keeps the list here for internal use.
 */
static int vms_mail_list_folders(int *UserData, struct DESC *FolderName)
  {
    char folder[FOLDERSIZE];

    if(FolderName == NULL) return;	/* Shouldn't happen */
    if(FolderName->length == 0) return;	/* End of list */

    strncpy(FoldersNames[FoldersNamesCount], FolderName->address, FolderName->length);
    FoldersNames[FoldersNamesCount][FolderName->length & 0xffff] = '\0';
    *FoldersNames[++FoldersNamesCount] = '\0';	/* End of list */
    return 1;
}

/* Parse flag list
 * Accepts: MAIL stream
 *	    flag list as a character string
 * Returns: flag command list
 */


short vms_mail_getflags (MAILSTREAM *stream, char *flag)
{
  char *t;
  short f = 0;
  short i,j;
  if (flag && *flag) {		/* no-op if no flag string */
				/* check if a list and make sure valid */
    if ((i = (*flag == '(')) ^ (flag[strlen (flag)-1] == ')')) {
      mm_log ("Bad flag list",ERROR);
      return NIL;
    }
				/* copy the flag string w/o list construct */
    strncpy (LOCAL->buf,flag+i,(j = strlen (flag) - (2*i)));
    LOCAL->buf[j] = '\0';
    t = ucase (LOCAL->buf);	/* uppercase only from now on */

    while (*t) {		/* parse the flags */
      if (*t == '\\') {		/* system flag? */
	switch (*++t) {		/* dispatch based on first character */
	case 'S':		/* possible \Seen flag */
          if (strncmp(t+1, "EEN", 3) == 0) i = fSEEN;
	  t += 4;		/* skip past flag name */
	  break;
	case 'D':		/* possible \Deleted flag */
          if (strncmp(t+1, "ELETED", 6) == 0) i = fDELETED;
	  t += 7;		/* skip past flag name */
	  break;
	case 'F':		/* possible \Flagged flag */
          if (strncmp(t+1, "LAGGED", 6) == 0) i = fFLAGGED;
	  t += 7;		/* skip past flag name */
	  break;
	case 'A':		/* possible \Answered flag */
          if (strncmp(t+1,"NSWERED", 7) == 0) i = fANSWERED;
	  t += 8;		/* skip past flag name */
	  break;
	default:		/* unknown */
	  i = 0;
	  break;
	}
				/* add flag to flags list */
	if (i && ((*t == '\0') || (*t++ == ' '))) f |= i;
	else {			/* bitch about bogus flag */
	  mm_log ("Unknown system flag",ERROR);
	  return NIL;
	}
      }
      else {			/* no user flags yet */
	mm_log ("Unknown flag",ERROR);
	return NIL;
      }
    }
  }
  return f;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Copy a message from the current message/folder to a new mailbox

   Author:
	Andy Harper, Kings College London, November 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int vms_mail_message_copy(long msgno, char *outmfile, char *outfolder)
  {
    static long MessageNum, FileCreated, FoldCreated;

    struct ITEM_LIST CopyFolderItem[] = {
	{ strlen(outfolder), MAIL$_MESSAGE_FOLDER,   outfolder,      NULL},	/* must be 1st,foldername */
	{ strlen(outmfile),  MAIL$_MESSAGE_FILENAME, outmfile,       NULL},	/* must be 2nd */
	{ sizeof(msgno),     MAIL$_MESSAGE_ID,       (char *)&msgno, NULL},	/* Message number to move */
	{ 0, 0, NULL, NULL } };
    
    struct ITEM_LIST CreatedItemList[] = {
	{ sizeof(FileCreated), MAIL$_MESSAGE_FILE_CREATED,   (char *) &FileCreated, 0},
	{ sizeof(FoldCreated), MAIL$_MESSAGE_FOLDER_CREATED, (char *) &FoldCreated, 0},
	{ 0, 0, NULL, NULL } };

    int status;

    status = mail$message_copy(&MessageContext, CopyFolderItem, CreatedItemList);
    if ((chkstat("mail$message_copy", 0, status) & 0x1) == 0) return NIL;

    if (FileCreated) mm_log("Created mailbox file", NIL);
    if (FoldCreated) mm_log("Created mailbox folder", NIL);

    return T;
  }

/* VMS/MAIL copy messages
 * Accepts: MAIL stream
 *	    mailbox copy vector
 *	    mailbox name
 * Returns: T if success, NIL if failed
 */

long vms_mail_copy_messages (MAILSTREAM *stream, char *mailbox)
  {
    long i;
    char outmfile[MAILFILESIZE],outfolder[FOLDERSIZE];


    if (vms_parse_mailbox(mailbox, outmfile, outfolder) == NIL)
      {
        mm_log("copy/move: invalid destination folder", ERROR);
        return NIL;
      }

    for (i=1; i <=stream->nmsgs; i++)
      {
        if (mail_elt(stream,i)->sequence)
           {
              if (vms_mail_message_copy(i,outmfile,outfolder))
                {
                   mm_log("move/copy: message copy failed", ERROR);
                   return NIL;
                }
           };
      }
    return T;
  }




/* VMS/MAIL update status string
 * Accepts: MAIL stream
 *	    message number
 *	    flag saying whether or not to sync
 */

void vms_mail_update_status (MAILSTREAM *stream, long msgno, long syncflag)
  {
	/*~~~~ Does nothing at preset. */
  }

/*
 | set the new number of messages. There is no automatic way...
 */
static void vms_set_new_mail_count(int NewCount)
{
  static int	UserContext;
  static short	MessagesCount;
  int		status;
  struct	ITEM_LIST MessagesCountItem[] = {
		{ sizeof(MessagesCount), MAIL$_USER_SET_NEW_MESSAGES,
		  (char *)&MessagesCount, NULL },
		{ 0, 0, NULL, NULL }};

  MessagesCount = (short) NewCount;
  UserContext = 0;
  status = mail$user_begin(&UserContext, NullItemList, NullItemList);
  if ((chkstat("mail$user_begin", 0, status) & 0x1) == 0) return;
  status = mail$user_set_info(&UserContext, MessagesCountItem, NullItemList);
  if ((chkstat("mail$user_set_info", 0, status) & 0x1) == 0) return;
  status = mail$user_end(&UserContext, NullItemList, NullItemList);
  if ((chkstat("mail$user_end", 0, status) & 0x1) == 0) return;
}

/*
 | If the folder is 'InboxName' then file all read and undeleted messages in
 | MAIL folder.
 */
static int vms_file_inbox_messages( MAILSTREAM *stream)
{
	int	i;
	MESSAGECACHE	message;	/* File message uses this... */
	char	sequence[256];		/* SetFlags uses strings not integers... */
        char mailfile[MAILFILESIZE], folder[FOLDERSIZE];

        vms_parse_mailbox(stream->mailbox, mailfile, folder);

/* Check whether we have something to do */
	if(strcmp(folder, InboxName) != 0)
		return;	/* Not 'InboxName' - ignore */


/* loop over all messages and check which are marked as SEEN and not as
   DELETED.
*/
	for(i = 1; i <= stream->nmsgs; i++)
	  {
  	    if(MessageTextCache[i - 1].flags & fSEEN)
	      {
		if((MessageTextCache[i - 1].flags & fDELETED) != fDELETED)
		  {
		    message.msgno = i;
		    vms_move_folder("MAIL", NULL, &message);
		    sprintf(sequence, "%d", i);
		    vms_mail_setflag (stream,sequence,"(\\DELETED)");	/* Delete it from 'InboxName' */
		  }
	      }
	  }

/* Now remove all deleted messages from 'InboxName'. We do not leave it to the close
   routine since the user might answer there with N and then we have two copies
   of the message: one in 'InboxName' and one in MAIL.
*/
	vms_mail_expunge (stream);
}

/*
 * VMS specific mail file open.
 *
 * Any previous context is saved (up to MAXSAVEPREV levels are permitted) and 
 * the new context set up in global variables.
 *
 */
static int __vms_mail_open(char *MailFile)
  {
    int status;
    struct ITEM_LIST MailFileList[] = {	/* Set the name of the file */
	{ strlen(MailFile), MAIL$_MAILFILE_NAME, MailFile, NULL },	/* Will fill values later */
	{ 0, 0, NULL, NULL}};

#ifdef MAILTRACE
    fprintf(stderr,"__vms_mail_open: Entered.. File %s\n", MailFile);
    fprintf(stderr,"__vms_mail_open: [%d] Saving mail context\n",PrevSP);
#endif
					/* Save current context */
    if (PrevSP >= MAXSAVEPREV)
      {
        fprintf(stderr,"__vms_mail_open: save context stack overflow (%d)\n", MAXSAVEPREV);
        return NIL;
      }
    PrevMailContext[PrevSP]        = MailContext;
    PrevMessageContext[PrevSP]     = MessageContext;
    PrevTotalMessagesCount[PrevSP] = TotalMessagesCount;
    PrevMessageCache[PrevSP]       = MessageTextCache;
    strcpy(PrevGlobalMailFile[PrevSP], GlobalMailFile);
    PrevSP++;

    MailContext = 0;
    MessageContext = 0;
    TotalMessagesCount = -1;
        
					/* Set up mailfile context */
    status = mail$mailfile_begin(&MailContext, NullItemList, NullItemList);
    chkstat("mail$mailfile_begin", 1, status);
    status = mail$mailfile_open(&MailContext, MailFileList, NullItemList);
    chkstat("mail$mailfile_open", 1, status);

    strcpy(GlobalMailFile, MailFile);

#ifdef MAILTRACE
    fprintf(stderr,"__vms_mail_open: exiting, mail ctx=0x%08x, message ctx=0x%08x\n", MailContext, MessageContext);
#endif
    return T;
  }

/* VMS specific mail file close. The context is taken from MailContext which is
 * a global variable.
 */
static void __vms_mail_close()
  {
    int status;
    struct ITEM_LIST CloseItem[] = {	/* Purge and reclaim during close */
	{ 0, MAIL$_MAILFILE_FULL_CLOSE, NULL, NULL},
	{ 0, 0, NULL, NULL } };

#ifdef MAILTRACE
    fprintf(stderr,"__vms_mail_close: Entered\n");
#endif
    if (MailContext != 0)
      {
        if(MessageContext != 0)
          {
#ifdef MAILTRACE
            fprintf(stderr,"__vms_mail_close: Terminating message context\n");
#endif
    	    status = mail$message_end(&MessageContext, NullItemList, NullItemList);
            chkstat("mail$message_end", 0, status);
          }

				/* Terminate the mail context */
#ifdef MAILTRACE
        fprintf(stderr,"__vms_mail_close: Terminating mail context (0x%08x)\n",MailContext);
#endif
        status = mail$mailfile_close(&MailContext, CloseItem, NullItemList);
        chkstat("mail$mailfile_close", 0, status);
        status = mail$mailfile_end(&MailContext, CloseItem, NullItemList);
        chkstat("mail$mailfile_end", 0, status);


				/* Restore any previous mail context */
	PrevSP--;
        if (PrevSP < 0)
          {
             fprintf(stderr,"__vms_mail_close: context stack underflow (%d)\n",PrevSP);
             exit(1);
          }

#ifdef MAILTRACE
        fprintf(stderr,"__vms_mail_close: [%d]: restoring previous context\n",PrevSP);
#endif
	MailContext        = PrevMailContext[PrevSP];
	MessageContext     = PrevMessageContext[PrevSP];
	TotalMessagesCount = PrevTotalMessagesCount[PrevSP];
        MessageTextCache   = PrevMessageCache[PrevSP];
	strcpy(GlobalMailFile, PrevGlobalMailFile[PrevSP]);
    }
#ifdef MAILTRACE
  fprintf(stderr,"__vms_mail_close: exiting, mail ctx=0x%08x, msg ctx=0x%08x\n", MailContext, MessageContext);
#endif
}

/* VMS/MAIL mail return internal message size in bytes. Since we get the size
 * in records we multiply it by MAXRECORDSIZE which is the maximal record length that
 * VMS/MAIL is willing to give us.
 *
 * Accepts: MAIL stream
 *	    message #
 * Returns: internal size of message
 */
unsigned long vms_mail_size (MAILSTREAM *stream, long m)
{
  MESSAGECACHE *elt = mail_elt(stream,m);
  int MessageSize, status;
  struct ITEM_LIST GetMessageInfoItem[] = {
	{ sizeof(m), MAIL$_MESSAGE_ID, (char *)&m, NULL },
	{ 0, 0, NULL, NULL }};
  struct ITEM_LIST AnswerMessageInfoItem[] = {
	{ sizeof(MessageSize), MAIL$_MESSAGE_SIZE, (char *)&MessageSize, NULL },
	{ 0, 0, NULL, NULL }};

  status = mail$message_get(&MessageContext, GetMessageInfoItem, AnswerMessageInfoItem);
  chkstat("mail$message_get", 1, status);

  if(MessageSize == 0) MessageSize = 1;	/* So all Malloc's won't fail */
  elt->data1 = elt->data2 = MessageSize * MAXRECORDSIZE;
  return (elt->data1);
}

/* VMS/MAIL mail validate mailbox
 * Accepts: mailbox name
 * Returns: our driver if name is valid, NIL otherwise
 */

DRIVER *vms_mail_valid (char *name)
{
  char mailfile[MAILFILESIZE], folder[FOLDERSIZE];
  return vms_parse_mailbox(name,mailfile,folder) ? &vmsmaildriver : NIL;
}



/* VMS/MAIL manipulate driver parameters
 * Accepts: function code
 *	    function-dependent value
 * Returns: function-dependent return value
 */

void *vms_mail_parameters (long function, void *value)
{
  return NIL;
}



/* VMS/MAIL mail find list of mailboxes
 * Accepts: mail stream
 *	    pattern to search
 */

void vms_mail_find (MAILSTREAM *stream, char *pat)
{
  vms_mail_find_all(stream, pat);	/* find all folders */
}


/* VMS/MAIL mail find list of bboards
 * Accepts: mail stream
 *	    pattern to search
 */

void vms_mail_find_bboards (MAILSTREAM *stream, char *pat)
{
  return;		/* no bboards supported on VMS */
}



/* VMS/MAIL mail find list of all bboards
 * Accepts: mail stream
 *	    pattern to search
 */

void vms_mail_find_all_bboards (MAILSTREAM *stream,char *pat)
{
   return;		/* Again, not supported at this stage */
}

/* VMS/MAIL mail find list of all mailboxes - in VMS'ish it is simply get
 *  the list of all folders.
 * Accepts: mail stream
 *	    pattern to search
 */

void vms_mail_find_all (MAILSTREAM *stream, char *pat)
{
  int status;
  char MailDirectory[MAILFILESIZE], folder[FOLDERSIZE];
  struct ITEM_LIST ListFoldersItem[] = {	/* get a list of all folders */
	{ sizeof(char *), MAIL$_MAILFILE_FOLDER_ROUTINE, (char *)list_folders, NULL },
	{ 0, 0, NULL, NULL } };


#ifdef MAILTRACE
        fprintf(stderr,"\nvms_mail_find_all: entered, pattern = %s\n\n", pat);
#endif

/* Check whether this is the postponed or interrupted mail. If so, check for
   the file. Must be synchronised with pine/OS_VMS.H
*/
  if((strstr(pat, "POSTPONED-MAIL") != NULL) ||
     (strstr(pat, "INTERRUPTED-MAIL") != NULL)) {
	FILE	*ifd;
	if((ifd = fopen(pat, "r")) != NULL) {
		mm_mailbox(pat);		/* Register it */
		fclose(ifd);
	}
	return;
  }

  if (vms_parse_mailbox(pat, MailDirectory, folder) == NIL)
    {
       mm_log("find_all: Invalid mailbox pattern", ERROR);
       return;
    }
						/* Initialize the pattern */
  pattern = pat;
						/* open the mail file */
  if (__vms_mail_open(MailDirectory) == NIL) return;
  status = mail$mailfile_info_file(&MailContext, ListFoldersItem, NullItemList);
  chkstat("mail$mailfile_info_file", 1, status);
  __vms_mail_close();			/* Close it */

#ifdef MAILTRACE
        fprintf(stderr,"\nvms_mail_find_all: exiting..\n\n");
#endif
}

/* VMS/MAIL mail subscribe to mailbox
 * Accepts: mail stream
 *	    mailbox to add to subscription list
 * Returns: T on success, NIL on failure
 */

long vms_mail_subscribe (MAILSTREAM *stream, char *mailbox)
{
  return NIL;		/* Only default mailbox on VMS */
}


/* VMS/MAIL mail unsubscribe to mailbox
 * Accepts: mail stream
 *	    mailbox to delete from subscription list
 * Returns: T on success, NIL on failure
 */

long vms_mail_unsubscribe (MAILSTREAM *stream, char *mailbox)
{
  return NIL;		/* Only default mailbox on VMS */
}


/* VMS/MAIL mail subscribe to bboard
 * Accepts: mail stream
 *	    bboard to add to subscription list
 * Returns: T on success, NIL on failure
 */

long vms_mail_subscribe_bboard (MAILSTREAM *stream, char *mailbox)
{
  return NIL;			/* never valid for VMS/MAIL */
}


/* VMS/MAIL mail unsubscribe to bboard
 * Accepts: mail stream
 *	    bboard to delete from subscription list
 * Returns: T on success, NIL on failure
 */

long vms_mail_unsubscribe_bboard (MAILSTREAM *stream, char *mailbox)
{
  return NIL;			/* never valid for VMS/MAIL */
}

/* VMS/MAIL mail create mailbox
 * Accepts: MAIL stream
 *	    mailbox name to create
 * Returns: T on success, NIL on failure
 */

long vms_mail_create (MAILSTREAM *stream, char *mailbox)
{
   mm_log("Create mailbox is not supported on VMS", NIL);
   return NIL;		/* We use only the default, remember? */
}


/* VMS/MAIL mail delete mailbox
 * Accepts: MAIL stream
 *	    mailbox name to delete
 * Returns: T on success, NIL on failure
 */

long vms_mail_delete (MAILSTREAM *stream, char *mailbox)
{
  mm_log("Delete mailbox is not supported on VMS", NIL);
  return NIL;
}


/* VMS/MAIL mail rename mailbox
 * Accepts: MAIL stream
 *	    old mailbox name
 *	    new mailbox name (or NIL for delete)
 * Returns: T on success, NIL on failure
 */

long vms_mail_rename (MAILSTREAM *stream, char *old, char *new)
{
   mm_log("Rename mailbox is not supported on VMS", NIL);
   return NIL;
}

/* VMS/MAIL mail open - open a "mailbox" which is a folder in VMS/MAIL.
 * Accepts: stream to open
 * Returns: stream on success, NIL on failure
 */

MAILSTREAM *vms_mail_open (MAILSTREAM *stream)
{
  long i, status;
  int fd,ld;
  char *s,*t,*k;
  char *MailboxName;
  char tmp[MAILTMPLEN];
  struct stat sbuf;
  struct ITEM_LIST MessageContextItem[] = { /* get the context for the folder */
	{ sizeof(MailContext), MAIL$_MESSAGE_FILE_CTX, (char *)&MailContext, NULL },
	{ 0, 0, NULL, NULL }};
  struct ITEM_LIST ListFoldersItem[] = {	/* get a list of all folders */
	{ sizeof(char *), MAIL$_MAILFILE_FOLDER_ROUTINE, (char *)vms_mail_list_folders, NULL },
	{ 0, 0, NULL, NULL } };
  char *p, *q, MailDirectory[MAILFILESIZE], folder[FOLDERSIZE];

  if (!stream) return &vmsmailproto;

#ifdef MAILTRACE
        fprintf(stderr,"\nvms_mail_open: Entered, mailbox '%s'\n\n", stream->mailbox);
#endif

  if (vms_parse_mailbox(stream->mailbox, MailDirectory, folder) == NIL)
    {
      mm_log("invalid mailbox format", ERROR);
      return NIL;
    }


				/* return prototype for OP_PROTOTYPE call */
  if (LOCAL) {			/* close old file if stream being recycled */
    vms_mail_close (stream);	/* dump and save the changes */
    stream->dtb = &vmsmaildriver;	/* reattach this driver */
    mail_free_cache (stream);	/* clean up cache */
  }


				/* force readonly if bboard */
  if (*stream->mailbox == '*') stream->rdonly = T;

/* Open the VMS file and select the folder we want */
  if (__vms_mail_open(MailDirectory) == NIL)
    {
       mm_log("no such folder", ERROR);
       return NIL;
    }

/* Get the list of folders so we can check whether the requested folder
   exists or not.
*/
#ifdef MAILTRACE
        fprintf(stderr,"\nvms_mail_open: creating folder list\n\n");
#endif

   FoldersNamesCount = 0;
   status = mail$mailfile_info_file(&MailContext, ListFoldersItem, NullItemList);
   chkstat("mail$mailfile_info_file", 1, status);

#ifdef MAILTRACE
        fprintf(stderr,"\nvms_mail_open: Searching for folder '%s'\n\n", folder);
#endif

   for(i = 0; *FoldersNames[i] != '\0'; i++)
	if(strcmp(FoldersNames[i], folder) == 0)
		break;

   if(*FoldersNames[i] == '\0') {	/* Not found */
#ifdef MAILTRACE
        fprintf(stderr,"\nvms_mail_open: folder '%s' not found\n\n", folder);
#endif
	__vms_mail_close();
	mm_log("No such folder", ERROR);
	return NIL;
   }

					/* Create folder and message context */
  MessageContext = 0;
  status = mail$message_begin(&MessageContext, MessageContextItem, NullItemList);
  chkstat("mail$message_begin", 1, status);

#ifdef MAILTRACE
        fprintf(stderr,"\nvms_mail_open: created message context 0x%08x\n\n", MessageContext);
#endif

  stream->dtb = &vmsmaildriver;	/* reattach this driver */
  stream->local = fs_get (sizeof (VMSMAILLOCAL));

				/* note if an INBOX or not */
  ucase(strcpy(tmp, stream->mailbox));
  LOCAL->inbox = (!strcmp (tmp, "INBOX") ||
                  !strcmp (folder, InboxName));

  LOCAL->filesize = 0;		/* initialize parsed file size */
  LOCAL->buf = (char *) fs_get (MAXMESSAGESIZE + 1);
  LOCAL->buflen = MAXMESSAGESIZE;
  stream->sequence++;		/* bump sequence number */
				/* parse mailbox */
  stream->nmsgs = stream->recent = 0;
  if (vms_mail_ping (stream))
    if(!stream->nmsgs)
	mm_log ("Mailbox is empty",(long) NIL);
  init_vms_mail_cache(stream->nmsgs);

/* Set the flags. If this is the 'InboxName'folder - All messages are
   defaulting to NEW. If this is another folder set all messages to seen.
*/
  if(strcmp(folder, InboxName) != 0) {
	sprintf(tmp, "1:%d", stream->nmsgs);
	vms_mail_setflag (stream, tmp, "(\\SEEN)");
  }
  for (i = 1; i < stream->nmsgs; i++) mail_elt (stream,i)->valid = T;


#ifdef MAILTRACE
  fprintf(stderr,"\nvms_mail_open: exiting\n\n");
#endif

  return stream;		/* return stream to caller */

}

/* VMS/MAIL mail close
 * Accepts: MAIL stream
 */

void vms_mail_close (MAILSTREAM *stream)
{
#ifdef MAILTRACE
  fprintf(stderr,"\nvms_mail_close: entered, mailbox '%s'\n", stream->mailbox);
#endif

					/* if the stream is open, close it */
  if (stream && MailContext)
    {
/*    vms_file_inbox_messages(stream); */		/* auto-file read msg*/
      clear_vms_mail_cache(stream->nmsgs);		/* clear cache */
      __vms_mail_close();				/* close file */
      if (LOCAL->buf) fs_give ((void **) &LOCAL->buf);	/* return local data*/
      fs_give ((void **) &stream->local);
      stream->dtb = NIL;				/* log out the DTB */
    }

#ifdef MAILTRACE
  fprintf(stderr,"\nvms_mail_close: exiting\n\n");
#endif
}

/* VMS/MAIL mail fetch fast information
 * Accepts: MAIL stream
 *	    sequence
 */

void vms_mail_fetchfast (MAILSTREAM *stream, char *sequence)
{
  long i;

				/* scan the sequence */
  if (mail_sequence (stream,sequence))
    for (i = 1; i <= stream->nmsgs; i++)
      if (mail_elt (stream,i)->sequence) {
         vms_mail_fetchstructure(stream,i,NIL);
      }

  return;
}


/* VMS/MAIL mail fetch flags
 * Accepts: MAIL stream
 *	    sequence
 */

void vms_mail_fetchflags (MAILSTREAM *stream, char *sequence)
{
  long i;

				/* ping mailbox, get new status for messages */
  if (mail_sequence (stream,sequence))
    for (i = 1; i <= stream->nmsgs; i++)
      if (mail_elt (stream,i)->sequence) {
	MESSAGECACHE *elt = mail_elt (stream,i);
	elt->deleted  = ((MessageTextCache[i - 1].flags & fDELETED)  ? T : NIL);
	elt->seen     = ((MessageTextCache[i - 1].flags & fSEEN)     ? T : NIL);
	elt->flagged  = ((MessageTextCache[i - 1].flags & fFLAGGED)  ? T : NIL);
	elt->answered = ((MessageTextCache[i - 1].flags & fANSWERED) ? T : NIL);
     }
}

/* VMS/MAIL mail fetch structure
 * Accepts: MAIL stream
 *	    message # to fetch
 *	    pointer to return body
 * Returns: envelope of this message, body returned in body value
 *
 * Fetches the "fast" information as well
 */
ENVELOPE *vms_mail_fetchstructure (MAILSTREAM *stream, long msgno, BODY **body)
{
  LONGCACHE *lelt;
  ENVELOPE **env;
  BODY **b;
  STRING bs;
  char *text, *hdr;
  unsigned long textsize = 0, hdrsize = 0;
  unsigned long hdrpos = 0;	/* Not implemented */
  unsigned long size = vms_mail_size (stream,msgno);
  unsigned long i = size;

  if (stream->scache) {		/* short cache */
    if (msgno != stream->msgno){/* flush old poop if a different message */
      mail_free_envelope (&stream->env);
      mail_free_body (&stream->body);
    }
    stream->msgno = msgno;
    env = &stream->env;		/* get pointers to envelope and body */
    b = &stream->body;
  }
  else {			/* long cache */
    lelt = mail_lelt (stream,msgno);
    env = &lelt->env;		/* get pointers to envelope and body */
    b = &lelt->body;
  }
  if ((body && !*b) || !*env) {	/* have the poop we need? */
    mail_free_envelope (env);	/* flush old envelope and body */
    mail_free_body (b);
    if (i > LOCAL->buflen) {	/* make sure enough buffer space */
      fs_give ((void **) &LOCAL->buf);
      LOCAL->buf = (char *) fs_get ((LOCAL->buflen = i) + 1);
    }

    vms_split_message(stream,msgno,size,&hdr,&hdrsize,&text,&textsize);

    INIT (&bs,mail_string,(void *) text,textsize);
				/* parse envelope and body */
    rfc822_parse_msg (env,body ? b : NIL,hdr,hdrsize,&bs,mylocalhost(),LOCAL->buf);
    (lelt->elt).rfc822_size = hdrsize + textsize;

/* Save the message's text in our buffer if it is not saved already */
    if(body) {
	if(MessageTextCache[msgno - 1].text == NIL) {	/* Not used yet */
		MessageTextCache[msgno - 1].size = textsize;
		MessageTextCache[msgno - 1].text = fs_get(textsize + 1);
		memcpy(MessageTextCache[msgno - 1].text, text, textsize + 1);
	}
    }

    fs_give ((void **) &text);
    fs_give ((void **) &hdr);
  }
  if (body) *body = *b;		/* return the body */
  return *env;			/* return the envelope */
}

/* VMS/MAIL mail fetch message header
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message header in RFC822 format
 */

char *vms_mail_fetchheader (MAILSTREAM *stream,long msgno)
{
  char	*hdr;
  unsigned long int hdrsize;
  unsigned long int size = vms_mail_size (stream,msgno);	/* Make this message current */

  vms_split_message(stream,msgno,size,&hdr,&hdrsize,NULL,NULL);
  strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,hdr,hdrsize);
  fs_give((void **) &hdr);
  return LOCAL->buf;
}




/* VMS/MAIL mail fetch message text (only)
	body only;
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message text in RFC822 format
 */

char *vms_mail_fetchtext (MAILSTREAM *stream, long msgno)
{
  char	*text;
  unsigned long textsize, hdrsize;
  unsigned long size = vms_mail_size (stream,msgno);


  vms_flag_seen_if_inbox(stream,msgno);
  mail_elt (stream,msgno)->seen = T;
				/* recalculate status */
  vms_mail_update_status (stream,msgno,T);
				/* get to text position */


				/* get to text position */
  vms_split_message(stream,msgno,size,NULL,NULL,&text,&textsize);
				/* copy the string */
  strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,text,textsize);
  fs_give((void **) &text);
  return LOCAL->buf;
}

/* VMS/MAIL fetch message body as a structure
 * Accepts: Mail stream
 *	    message # to fetch
 *	    section specifier
 *	    pointer to length
 * Returns: pointer to section of message body
 */
char *vms_mail_fetchbody (MAILSTREAM *stream, long m, char *s, unsigned long *len)
{
  BODY *b;
  PART *pt;
  char *t;
  ENVELOPE *env;
  unsigned long i;
  unsigned long offset = 0;
  MESSAGECACHE *elt = mail_elt (stream,m);

  vms_flag_seen_if_inbox(stream,m);

				/* make sure have a body */
  env = vms_mail_fetchstructure (stream,m,&b);
  if (!(env && b && s && *s &&
	((i = strtol (s,&s,10)) > 0))) return NIL;
  do {				/* until find desired body part */
				/* multipart content? */
    if (b->type == TYPEMULTIPART) {
      pt = b->contents.part;	/* yes, find desired part */
      while (--i && (pt = pt->next));
      if (!pt) return NIL;	/* bad specifier */
				/* note new body, check valid nesting */
      if (((b = &pt->body)->type == TYPEMULTIPART) && !*s) return NIL;
      offset = pt->offset;	/* get new offset */
    }
    else if (i != 1) return NIL;/* otherwise must be section 1 */
				/* need to go down further? */
    if (i = *s) switch (b->type) {
    case TYPEMESSAGE:		/* embedded message, calculate new base */
      offset = b->contents.msg.offset;
      b = b->contents.msg.body;	/* get its body, drop into multipart case */
    case TYPEMULTIPART:		/* multipart, get next section */
      if ((*s++ == '.') && (i = strtol (s,&s,10)) > 0) break;
    default:			/* bogus subpart specification */
      return NIL;
    }
  } while (i);
				/* lose if body bogus */
  if ((!b) || b->type == TYPEMULTIPART) return NIL;
				/* move to that place in the data */

/*  lseek (LOCAL->fd,offset,L_SET); */
  t = (char *) fs_get (1 + b->size.ibytes);

  strncpy(t, &((MessageTextCache[m - 1].text)[offset]), b->size.ibytes);

/*  read (LOCAL->fd,t,b->size.ibytes); */

  rfc822_contents(&LOCAL->buf,&LOCAL->buflen,len,t,b->size.ibytes,b->encoding);
  fs_give ((void **) &t);	/* flush readin buffer */
  return LOCAL->buf;
}

/* VMS/MAIL mail set flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void vms_mail_setflag (MAILSTREAM *stream, char *sequence, char *flag)
{
  int	i;
  MESSAGECACHE *elt;

  short f = vms_mail_getflags (stream,flag);	/* Parse the list */
  if (!f) return;		/* no-op if no flags to modify */
				/* get sequence and loop on it */
  if (mail_sequence (stream,sequence)) for (i = 1; i <= stream->nmsgs; i++)
    if ((elt = mail_elt (stream,i))->sequence) {
				/* set all requested flags */
      if (f&fSEEN)    {elt->seen     = T; MessageTextCache[i - 1].flags |= fSEEN;}
      if (f&fDELETED) {elt->deleted  = T; MessageTextCache[i - 1].flags |= fDELETED;}
      if (f&fFLAGGED)  elt->flagged  = T;
      if (f&fANSWERED) elt->answered = T;
    }
}




/* VMS/MAIL mail clear flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void vms_mail_clearflag (MAILSTREAM *stream, char *sequence, char *flag)
{
  int	i;
  MESSAGECACHE *elt;

  short f = vms_mail_getflags (stream,flag);	/* Parse the list */
  if (!f) return;		/* no-op if no flags to modify */
				/* get sequence and loop on it */
  if (mail_sequence (stream,sequence)) for (i = 1; i <= stream->nmsgs; i++)
    if ((elt = mail_elt (stream,i))->sequence) {
				/* set all requested flags */
      if (f&fSEEN)    {elt->seen     = NIL;MessageTextCache[i - 1].flags &= ~fSEEN;}
      if (f&fDELETED) {elt->deleted  = NIL;MessageTextCache[i - 1].flags &= ~fDELETED;}
      if (f&fFLAGGED)  elt->flagged  = NIL;
      if (f&fANSWERED) elt->answered = NIL;
    }
}

/* VMS/MAIL mail search for messages
 * Accepts: MAIL stream
 *	    search criteria
 */

void vms_mail_search (MAILSTREAM *stream, char *criteria)
{
  long i,n;
  char *d;
  search_t f;
				/* initially all searched */
  for (i = 1; i <= stream->nmsgs; ++i) mail_elt (stream,i)->searched = T;
				/* get first criterion */
  if (criteria && (criteria = strtok (criteria," "))) {
				/* for each criterion */
    for (; criteria; (criteria = strtok (NIL," "))) {
      f = NIL; d = NIL; n = 0;	/* init then scan the criterion */
      switch (*ucase (criteria)) {
      case 'A':			/* possible ALL, ANSWERED */
	if (!strcmp (criteria+1,"LL")) f = vms_mail_search_all;
	else if (!strcmp (criteria+1,"NSWERED")) f = vms_mail_search_answered;
	break;
      case 'B':			/* possible BCC, BEFORE, BODY */
	if (!strcmp (criteria+1,"CC"))
	  f = vms_mail_search_string (vms_mail_search_bcc,&d,&n);
	else if (!strcmp (criteria+1,"EFORE"))
	  f = vms_mail_search_date (vms_mail_search_before,&n);
	else if (!strcmp (criteria+1,"ODY"))
	  f = vms_mail_search_string (vms_mail_search_body,&d,&n);
	break;
      case 'C':			/* possible CC */
	if (!strcmp (criteria+1,"C"))
	  f = vms_mail_search_string (vms_mail_search_cc,&d,&n);
	break;
      case 'D':			/* possible DELETED */
	if (!strcmp (criteria+1,"ELETED")) f = vms_mail_search_deleted;
	break;
      case 'F':			/* possible FLAGGED, FROM */
	if (!strcmp (criteria+1,"LAGGED")) f = vms_mail_search_flagged;
	else if (!strcmp (criteria+1,"ROM"))
	  f = vms_mail_search_string (vms_mail_search_from,&d,&n);
	break;
      case 'K':			/* possible KEYWORD */
	if (!strcmp (criteria+1,"EYWORD"))
	  f = vms_mail_search_flag (vms_mail_search_keyword,&n,stream);
	break;
      case 'N':			/* possible NEW */
	if (!strcmp (criteria+1,"EW")) f = vms_mail_search_new;
	break;
      case 'O':			/* possible OLD, ON */
	if (!strcmp (criteria+1,"LD")) f = vms_mail_search_old;
	else if (!strcmp (criteria+1,"N"))
	  f = vms_mail_search_date (vms_mail_search_on,&n);
	break;
      case 'R':			/* possible RECENT */
	if (!strcmp (criteria+1,"ECENT")) f = vms_mail_search_recent;
	break;
      case 'S':			/* possible SEEN, SINCE, SUBJECT */
	if (!strcmp (criteria+1,"EEN")) f = vms_mail_search_seen;
	else if (!strcmp (criteria+1,"INCE"))
	  f = vms_mail_search_date (vms_mail_search_since,&n);
	else if (!strcmp (criteria+1,"UBJECT"))
	  f = vms_mail_search_string (vms_mail_search_subject,&d,&n);
	break;
      case 'T':			/* possible TEXT, TO */
	if (!strcmp (criteria+1,"EXT"))
	  f = vms_mail_search_string (vms_mail_search_text,&d,&n);
	else if (!strcmp (criteria+1,"O"))
	  f = vms_mail_search_string (vms_mail_search_to,&d,&n);
	break;
      case 'U':			/* possible UN* */
	if (criteria[1] == 'N') {
	  if (!strcmp (criteria+2,"ANSWERED")) f = vms_mail_search_unanswered;
	  else if (!strcmp (criteria+2,"DELETED")) f = vms_mail_search_undeleted;
	  else if (!strcmp (criteria+2,"FLAGGED")) f = vms_mail_search_unflagged;
	  else if (!strcmp (criteria+2,"KEYWORD"))
	    f = vms_mail_search_flag (vms_mail_search_unkeyword,&n,stream);
	  else if (!strcmp (criteria+2,"SEEN")) f = vms_mail_search_unseen;
	}
	break;
      default:			/* we will barf below */
	break;
      }
      if (!f) {			/* if can't determine any criteria */
	sprintf (LOCAL->buf,"Unknown search criterion: %.80s",criteria);
	mm_log (LOCAL->buf,ERROR);
	return;
      }
				/* run the search criterion */
      for (i = 1; i <= stream->nmsgs; ++i)
	if (mail_elt (stream,i)->searched && !(*f) (stream,i,d,n))
	  mail_elt (stream,i)->searched = NIL;
    }
				/* report search results to main program */
    for (i = 1; i <= stream->nmsgs; ++i)
      if (mail_elt (stream,i)->searched) mail_searched (stream,i);
  }
}

/* VMS/MAIL mail ping mailbox - find how many messages are in this folder.
 * Do it only when openning the folder. TotalMessagesCount remembers the number
 * and we do not check again in the middle to not confuse the caching system.
 * Accepts: MAIL stream
 * Returns: T if stream still alive, NIL if not
 */
long vms_mail_ping (MAILSTREAM *stream)
{
  long r = NIL;
  int	status;
  static unsigned int MessagesCount;
  struct ITEM_LIST FolderItem[] = {	/* Folder's name */
	{ 0, MAIL$_MESSAGE_FOLDER, NULL, NULL }, /* Must be the first item!!! */
	{ 0, 0, NULL, NULL }};
  struct ITEM_LIST MessagesCountItem[] = { /* How many mesages are there */
	{ sizeof(MessagesCount), MAIL$_MESSAGE_SELECTED,(char *)&MessagesCount, NULL },
	{ 0, 0, NULL, NULL }};
  char mailfile[MAILFILESIZE], folder[FOLDERSIZE];

  vms_parse_mailbox(stream->mailbox, mailfile, folder);

  if(TotalMessagesCount == -1) {	/* Not checked yet */
	FolderItem[0].buffer = folder;
	FolderItem[0].length = strlen(folder);

	status = mail$message_select(&MessageContext, FolderItem, MessagesCountItem);
        if ((chkstat("mail$message_select", 0, status) & 0x1) == 0) return NIL;
	TotalMessagesCount = MessagesCount;
   }
   else	MessagesCount = TotalMessagesCount;	/* Ignore mails added in the middle */

  return (stream->nmsgs = MessagesCount) != 0 ? T : NIL;
}


/* VMS/MAIL mail check mailbox (too)
	reparses status too;
 * Accepts: MAIL stream
 */

void vms_mail_check (MAILSTREAM *stream)
{
  mm_log ("Check completed",(long) NIL);
}

/* VMS/MAIL mail expunge mailbox - Delete the messages that are flagged as
 * deleted.
 * Accepts: MAIL stream
 */

void vms_mail_expunge (MAILSTREAM *stream)
{
  static int MessageNum;		/* Message number to delete */
  struct ITEM_LIST DeleteMessageItem[] = {
	{ sizeof(MessageNum), MAIL$_MESSAGE_ID, (char *)&MessageNum, NULL },
	{ 0, 0, NULL, NULL }};
  int i, status, MessagesDeleted = 0;	/* How many messages were deleted */

  char mailfile[MAILFILESIZE], folder[FOLDERSIZE];

  vms_parse_mailbox(stream->mailbox, mailfile, folder);

  for (i = 1; i <= stream->nmsgs; ++i)
    {
      if(MessageTextCache[i - 1].flags & fDELETED)
        {
  	  MessagesDeleted++;

/* In 'InboxName' we are called twice, so mark which message was deleted and do
   not try deleting it again.
*/
  	  if((MessageTextCache[i - 1].flags & fWAS_DELETED) != fWAS_DELETED)
	    {
		MessageNum = i;	/* Assign it at the descriptor */
		status = mail$message_delete(&MessageContext, &DeleteMessageItem, NullItem);
                chkstat("mail$message_delete", 1, status);
		MessageTextCache[i - 1].flags |= fWAS_DELETED;
   	    }
        }
    }

  if((MessagesDeleted) &&		/* If Some messages were removed */
     (strcmp(folder, InboxName) == 0))	/* .... And it is 'InboxName'    */
	vms_set_new_mail_count(stream->nmsgs - MessagesDeleted);	/* Set the new count */
}

/* VMS/MAIL mail copy message(s)
	s;
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 * Returns: T if success, NIL if failed
 */

long vms_mail_copy (MAILSTREAM *stream, char *sequence, char *mailbox)
{
				/* copy the messages */
  return (mail_sequence (stream,sequence)) ?
    vms_mail_copy_messages (stream,mailbox) : NIL;
}


/* VMS/MAIL mail move message(s)
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 * Returns: T if success, NIL if failed
 */

long vms_mail_move (MAILSTREAM *stream, char *sequence, char *mailbox)
{
  long i;
  MESSAGECACHE *elt;
  if (!(mail_sequence (stream,sequence) &&
	vms_mail_copy_messages (stream,mailbox))) return NIL;

				/* delete all requested messages */
  for (i = 1; i <= stream->nmsgs; i++)
    if ((elt = mail_elt (stream,i))->sequence) {
      elt->deleted = T;		/* mark message deleted */
				/* recalculate status */
      vms_mail_update_status (stream,i,NIL);
    }
				/* make sure the update takes */
  return LONGT;
}

/* VMS/MAIL mail append message from stringstruct
 * Accepts: MAIL stream
 *	    destination mailbox
 *	    stringstruct of messages to append
 * Returns: T if append successful, else NIL
 */

long vms_mail_append (MAILSTREAM *stream, char *mailbox, STRING *message)
{
  return NIL;	/* Do nothing */
}

/* VMS/MAIL garbage collect stream
 * Accepts: Mail stream
 *	    garbage collection flags
 */

void vms_mail_gc (MAILSTREAM *stream, long gcflags)
{
  /* nothing here for now */
}

/* Search support routines
 * Accepts: MAIL stream
 *	    message number
 *	    pointer to additional data
 * Returns: T if search matches, else NIL
 */


char vms_mail_search_all (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return T;			/* ALL always succeeds */
}


char vms_mail_search_answered (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->answered ? T : NIL;
}


char vms_mail_search_deleted (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->deleted ? T : NIL;
}


char vms_mail_search_flagged (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->flagged ? T : NIL;
}


char vms_mail_search_keyword (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->user_flags & n ? T : NIL;
}


char vms_mail_search_new (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (elt->recent && !elt->seen) ? T : NIL;
}

char vms_mail_search_old (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->recent ? NIL : T;
}


char vms_mail_search_recent (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->recent ? T : NIL;
}


char vms_mail_search_seen (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->seen ? T : NIL;
}


char vms_mail_search_unanswered (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->answered ? NIL : T;
}


char vms_mail_search_undeleted (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->deleted ? NIL : T;
}


char vms_mail_search_unflagged (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->flagged ? NIL : T;
}


char vms_mail_search_unkeyword (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->user_flags & n ? NIL : T;
}


char vms_mail_search_unseen (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  return mail_elt (stream,msgno)->seen ? NIL : T;
}



char vms_mail_search_before (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (char) ((long)((elt->year << 9) + (elt->month << 5) + elt->day) < n);
}


char vms_mail_search_on (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (char) (((elt->year << 9) + (elt->month << 5) + elt->day) == n);
}


char vms_mail_search_since (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
				/* everybody interprets "since" as .GE. */
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (char) ((long)((elt->year << 9) + (elt->month << 5) + elt->day) >= n);
}

char vms_mail_search_body (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  char ret;
  char *s = vms_mail_fetchtext (stream,msgno);	/* get the text in LOCAL->buf */

  ret = search (s,strlen(s),d,n);/* do the search */
  return ret;			/* return search value */
}

char vms_mail_search_subject (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  char *s = vms_mail_fetchstructure (stream,msgno,NIL)->subject;
  return s ? search (s,(long) strlen (s),d,n) : NIL;
}


char vms_mail_search_text (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  char ret;
  char *s = vms_mail_fetchheader (stream,msgno);	/* Get the header in LOCAL->buf */

  ret = search (s,strlen(s),d,n);  /* First, search at the header */
  if(!ret)	/* Not found at header, search at the body */
	  ret = vms_mail_search_body (stream,msgno,d,n);
  return ret;			/* return search value */
}



char vms_mail_search_bcc (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->bcc;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}


char vms_mail_search_cc (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->cc;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}


char vms_mail_search_from (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->from;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}


char vms_mail_search_to (stream,msgno,d,n)
	MAILSTREAM *stream;
	long msgno;
	char *d;
	long n;
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->to;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}

/* Search parsers */


/* Parse a date
 * Accepts: function to return
 *	    pointer to date integer to return
 * Returns: function to return
 */

search_t vms_mail_search_date (f,n)
	search_t f;
	long *n;
{
  long i;
  char *s;
  MESSAGECACHE elt;
				/* parse the date and return fn if OK */
  return (vms_mail_search_string (f,&s,&i) && mail_parse_date (&elt,s) &&
	  (*n = (elt.year << 9) + (elt.month << 5) + elt.day)) ? f : NIL;
}

/* Parse a flag
 * Accepts: function to return
 *	    pointer to keyword integer to return
 *	    MAIL stream
 * Returns: function to return
 */

search_t vms_mail_search_flag (f,n,stream)
	search_t f;
	long *n;
	MAILSTREAM *stream;
{
  short i;
  char *s,*t;
  if (t = strtok (NIL," ")) {	/* get a keyword */
    ucase (t);			/* get uppercase form of flag */
    for (i = 0; i < NUSERFLAGS && (s = stream->user_flags[i]); ++i)
      if (!strcmp (t,ucase (strcpy (LOCAL->buf,s))) && (*n = 1 << i)) return f;
  }
  return NIL;			/* couldn't find keyword */
}

/* Parse a string
 * Accepts: function to return
 *	    pointer to string to return
 *	    pointer to string length to return
 * Returns: function to return
 */


search_t vms_mail_search_string (f,d,n)
	search_t f;
	char **d;
	long *n;
{
  char *end = " ";
  char *c = strtok (NIL,"");	/* remainder of criteria */
  if (!c) return NIL;		/* missing argument */
  switch (*c) {			/* see what the argument is */
  case '{':			/* literal string */
    *n = strtol (c+1,d,10);	/* get its length */
    if ((*(*d)++ == '}') && (*(*d)++ == '\015') && (*(*d)++ == '\012') &&
	(!(*(c = *d + *n)) || (*c == ' '))) {
      char e = *--c;
      *c = DELIM;		/* make sure not a space */
      strtok (c," ");		/* reset the strtok mechanism */
      *c = e;			/* put character back */
      break;
    }
  case '\0':			/* catch bogons */
  case ' ':
    return NIL;
  case '"':			/* quoted string */
    if (strchr (c+1,'"')) end = "\"";
    else return NIL;
  default:			/* atomic string */
    if (*d = strtok (c,end)) *n = strlen (*d);
    else return NIL;
    break;
  }
  return f;
}


/*
 | Copy a message from the current folder to another one. Do not mark the
 | original message for deletion as we do not have the Stream available...
 | Furthermore, the caller will set the D flag for it.
 |
 | Returns: 0 - some error.
 |          1 - success.
 */
int vms_move_folder(char *folder, void *context, MESSAGECACHE *message)
  {
    char MailFile[MAILFILESIZE], FolderName[FOLDERSIZE];

    if ( vms_parse_mailbox(folder,MailFile,FolderName) == NIL)
      {
         mm_log("copy/move: invalid output folder", ERROR);
         return 0;
      }

    return vms_mail_message_copy(message->msgno, MailFile, FolderName);
}

static unsigned int SendContext = 0;


/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Purpose:
	Check error status, cancel the current send and return pointer to
	suitable error message.

     Author:
	Andy Harper, Kings College London, Oct 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static char *chksndstat(char *message, int status)
  {
     static char errmess[MAXRECORDSIZE];
     struct ITEM_LIST NullItemList[] = {{ 0, 0, NULL, NULL }};	/* Null list */

     if ((status & 0x1) == 0)
       {
         sprintf(errmess,"%s: failed with status 0x%08x", message, status);
         if (SendContext != 0) mail$send_end(&SendContext, NullItemList, NullItemList);
         return errmess;
       }

     return NULL;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 | Get the message's text filename, the addresses and send it using the
 | selected mail protocol.
 | Returns NULL in case of success, or a pointer to the error message.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static int MessageSentCount;	/* Set by the error/success routine */

char *vms_mail_send(char *filename, ADDRESS *from, ADDRESS *to, ADDRESS *cc, char *subject)
  {
    ADDRESS *ap;
    static short int typesu = MAIL$_TO, typecc = MAIL$_CC;
    int status, AddressesCount, success_send_result();
    FILE *ifd;
    static char line[MAXRECORDSIZE];	/* used also to return error messages to caller */
    char *err,*p, MailProtocolUsed[256];	/* The prefix name */

    struct ITEM_LIST ToItem[] = {
	{ 0, MAIL$_SEND_USERNAME, line, NULL },
	{ sizeof(typesu), MAIL$_SEND_USERNAME_TYPE, (char *)&typesu, 0},
	{ 0, 0, NULL, NULL }};
    struct ITEM_LIST CcItem[] = {
	{ 0, MAIL$_SEND_USERNAME, line, NULL },
	{ sizeof(typecc), MAIL$_SEND_USERNAME_TYPE, (char *)&typecc, 0},
	{ 0, 0, NULL, NULL }};
    struct ITEM_LIST	FromItem[] = {
	{ 0, MAIL$_SEND_FROM_LINE, line, NULL },
	{ 0, 0, NULL, NULL }};
    struct ITEM_LIST	SubjectItem[] = {
	{ 0, MAIL$_SEND_SUBJECT, line, NULL },
	{ 0, 0, NULL, NULL }};
    struct ITEM_LIST	RecordItem[] = {
	{ 0, MAIL$_SEND_RECORD, line, NULL },
	{ 0, 0, NULL, NULL }};
    struct ITEM_LIST	FileItem[] = {
	{ 0, MAIL$_SEND_FILENAME, line, NULL },
	{ 0, 0, NULL, NULL }};
    struct ITEM_LIST	SendItem[] = {
	{ sizeof (int *), MAIL$_SEND_SUCCESS_ENTRY, (char *)success_send_result, NULL },
	{ 0, MAIL$_NOSIGNAL, NULL, NULL},
	{ 0, 0, NULL, NULL }};

/* Get the foreign protocol to use */
    if((p = getenv("PINE_MAIL_PROTOCOL")) != NULL)
	strcpy(MailProtocolUsed, p);

    else if ((p = getenv("MAIL$INTERNET_TRANSPORT")) != NULL)
       	strcpy(MailProtocolUsed, p);

    else /* No foreign protocol - can't work */
	return "No foreign mail transport protocol defined";

/* Create mail context */
    SendContext = 0;
    status = mail$send_begin(&SendContext, NullItemList, NullItemList);
    if ( (err = chksndstat("mail$send_begin", status)) != NULL) return err;

/* Create the sender's address */
/* Use the system default FROM: address (to avoid privileges)
    sprintf(line, "%s <%s@%s>", (from->personal ? from->personal : ""),
	from->mailbox, from->host);
    FromItem[0].length = strlen(line);
    status = mail$send_add_attribute(&SendContext, FromItem, NullItemList);
    if ( (err = chksndstat("mail$send_add_attribute", status)) != NULL) return err;
*/

/* Now the subject */
    if(subject) {
	strcpy(line, subject);
	SubjectItem[0].length = strlen(line);
	status = mail$send_add_attribute(&SendContext, SubjectItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_attribute", status)) != NULL) return err;
    }

/* now add the TO addresses */
    AddressesCount = 0;
    ap = to;
    while(ap != NULL) {
	sprintf(line, "%s%%\"%s@%s\"", MailProtocolUsed, ap->mailbox, ap->host);
	ToItem[0].length = strlen(line);
	status = mail$send_add_address(&SendContext, ToItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_address", status)) != NULL) return err;
	ap = ap->next;
	AddressesCount++;
    }

/* Same for CC */
    ap = cc;
    while(ap != NULL) {
	sprintf(line, "%s%%\"%s@%s\"", MailProtocolUsed, ap->mailbox, ap->host);
	CcItem[0].length = strlen(line);
	status = mail$send_add_address(&SendContext, CcItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_address", status)) != NULL) return err;
	ap = ap->next;
	AddressesCount++;
    }

/* Open the message's file */
    if((ifd = fopen(filename, "r")) == NULL) {
	mail$send_end(&SendContext, NullItemList, NullItemList);
	sprintf(line, "Can't open '%s'", filename);
	return line;
    }

/* Now copy it. VMS/MAIL can accept up to MAXRECORDSIZE characters long lines */

/* --- Method 1: Add each record, one at a time
    while(fgets(line, (sizeof(line) < MAXRECORDSIZE ? sizeof(line) - 1 : MAXRECORDSIZE), ifd) != NULL) {
	char *p = strchr(line, '\n');
	if(p) *p = '\0';
	RecordItem[0].length = strlen(line);
	status = mail$send_add_bodypart(&SendContext, RecordItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_bodypart", status)) != NULL) return err;
    }
*/
    fclose(ifd);

/* --- Method 2: Just add the filename */
    strcpy(line,filename);
    FileItem[0].length = strlen(line);
    status = mail$send_add_bodypart(&SendContext, FileItem, NullItemList);
    if ( (err = chksndstat("mail$send_add_bodypart", status)) != NULL) return err;



/* OK, done - actually send the message. This routine will call via AST a routine
   that will count how many were sent ok.
*/
    MessageSentCount = 0;	/* None set yet */
    status = mail$send_message(&SendContext, SendItem, NullItemList);
    if ( (err = chksndstat("mail$send_message", status)) != NULL) return err;

    status = mail$send_end(&SendContext, NullItemList, NullItemList);
    if ( (err = chksndstat("mail$send_end", status)) != NULL) return err;

    if(MessageSentCount < AddressesCount)
	return "Error at final stages of delivery";

    return NULL;	/* Success */
}




/*
 | this routine is called only if the message was sent ok. We count the number
 | of messages sent ok.
 */
int success_send_result(struct ITEM_LIST *recipient, int *sigarray, int userdata)
  {
    MessageSentCount++;	/* OK */
  }
